@@ -12,12 +12,15 @@
 #include <libxml/xmlstring.h>
 #include <librevenge-stream/librevenge-stream.h>
 #include <boost/algorithm/string.hpp>
+#include <boost/shared_ptr.hpp>
 #include "VSDXParser.h"
 #include "libvisio_utils.h"
+#include "libvisio_xml.h"
 #include "VSDContentCollector.h"
 #include "VSDStylesCollector.h"
 #include "VSDXMLHelper.h"
 #include "VSDXMLTokenMap.h"
+#include "VSDXMetaData.h"
 
 namespace
 {
@@ -59,50 +62,46 @@ libvisio::VSDXParser::~VSDXParser()
 {
 }
 
-bool libvisio::VSDXParser::parseMain()
+bool libvisio::VSDXParser::parseMain() try
 {
   if (!m_input || !m_input->isStructured())
     return false;
 
-  librevenge::RVNGInputStream *tmpInput = 0;
-  try
-  {
-    tmpInput = m_input->getSubStreamByName("_rels/.rels");
-    if (!tmpInput)
-      return false;
-
-    libvisio::VSDXRelationships rootRels(tmpInput);
-    delete tmpInput;
-
-    // Check whether the relationship points to a Visio document stream
-    const libvisio::VSDXRelationship *rel = rootRels.getRelationshipByType("http://schemas.microsoft.com/visio/2010/relationships/document");
-    if (!rel)
-      return false;
-
-    std::vector<std::map<unsigned, XForm> > groupXFormsSequence;
-    std::vector<std::map<unsigned, unsigned> > groupMembershipsSequence;
-    std::vector<std::list<unsigned> > documentPageShapeOrders;
-
-    VSDStylesCollector stylesCollector(groupXFormsSequence, groupMembershipsSequence, documentPageShapeOrders);
-    m_collector = &stylesCollector;
-    if (!parseDocument(m_input, rel->getTarget().c_str()))
-      return false;
-
-    VSDStyles styles = stylesCollector.getStyleSheets();
-
-    VSDContentCollector contentCollector(m_painter, groupXFormsSequence, groupMembershipsSequence, documentPageShapeOrders, styles, m_stencils);
-    m_collector = &contentCollector;
-    if (!parseDocument(m_input, rel->getTarget().c_str()))
-      return false;
-
-    return true;
-  }
-  catch (...)
-  {
-    if (tmpInput)
-      delete tmpInput;
+  RVNGInputStreamPtr_t tmpInput;
+  tmpInput.reset(m_input->getSubStreamByName("_rels/.rels"));
+  if (!tmpInput)
     return false;
-  }
+
+  libvisio::VSDXRelationships rootRels(tmpInput.get());
+
+  // Check whether the relationship points to a Visio document stream
+  const libvisio::VSDXRelationship *rel = rootRels.getRelationshipByType("http://schemas.microsoft.com/visio/2010/relationships/document");
+  if (!rel)
+    return false;
+
+  std::vector<std::map<unsigned, XForm> > groupXFormsSequence;
+  std::vector<std::map<unsigned, unsigned> > groupMembershipsSequence;
+  std::vector<std::list<unsigned> > documentPageShapeOrders;
+
+  VSDStylesCollector stylesCollector(groupXFormsSequence, groupMembershipsSequence, documentPageShapeOrders);
+  m_collector = &stylesCollector;
+  if (!parseDocument(m_input, rel->getTarget().c_str()))
+    return false;
+
+  VSDStyles styles = stylesCollector.getStyleSheets();
+
+  VSDContentCollector contentCollector(m_painter, groupXFormsSequence, groupMembershipsSequence, documentPageShapeOrders, styles, m_stencils);
+  m_collector = &contentCollector;
+  parseMetaData(m_input, rootRels);
+
+  if (!parseDocument(m_input, rel->getTarget().c_str()))
+    return false;
+
+  return true;
+}
+catch (...)
+{
+  return false;
 }
 
 bool libvisio::VSDXParser::extractStencils()
@@ -118,15 +117,13 @@ bool libvisio::VSDXParser::parseDocument(librevenge::RVNGInputStream *input, con
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!input->isStructured())
     return false;
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!stream)
     return false;
-  librevenge::RVNGInputStream *relStream = input->getSubStreamByName(getRelationshipsForTarget(name).c_str());
+  const RVNGInputStreamPtr_t relStream(input->getSubStreamByName(getRelationshipsForTarget(name).c_str()));
   input->seek(0, librevenge::RVNG_SEEK_SET);
-  VSDXRelationships rels(relStream);
-  if (relStream)
-    delete relStream;
+  VSDXRelationships rels(relStream.get());
   rels.rebaseTargets(getTargetBaseDirectory(name).c_str());
 
   const VSDXRelationship *rel = rels.getRelationshipByType("http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme");
@@ -139,7 +136,7 @@ bool libvisio::VSDXParser::parseDocument(librevenge::RVNGInputStream *input, con
     input->seek(0, librevenge::RVNG_SEEK_SET);
   }
 
-  processXmlDocument(stream, rels);
+  processXmlDocument(stream.get(), rels);
 
   rel = rels.getRelationshipByType("http://schemas.microsoft.com/visio/2010/relationships/masters");
   if (rel)
@@ -161,8 +158,6 @@ bool libvisio::VSDXParser::parseDocument(librevenge::RVNGInputStream *input, con
     input->seek(0, librevenge::RVNG_SEEK_SET);
   }
 
-  if (stream)
-    delete stream;
   return true;
 }
 
@@ -173,19 +168,16 @@ bool libvisio::VSDXParser::parseMasters(librevenge::RVNGInputStream *input, cons
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!input->isStructured())
     return false;
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   if (!stream)
     return false;
-  librevenge::RVNGInputStream *relStream = input->getSubStreamByName(getRelationshipsForTarget(name).c_str());
+  const RVNGInputStreamPtr_t relStream(input->getSubStreamByName(getRelationshipsForTarget(name).c_str()));
   input->seek(0, librevenge::RVNG_SEEK_SET);
-  VSDXRelationships rels(relStream);
-  if (relStream)
-    delete relStream;
+  VSDXRelationships rels(relStream.get());
   rels.rebaseTargets(getTargetBaseDirectory(name).c_str());
 
-  processXmlDocument(stream, rels);
+  processXmlDocument(stream.get(), rels);
 
-  delete stream;
   return true;
 }
 
@@ -196,19 +188,16 @@ bool libvisio::VSDXParser::parseMaster(librevenge::RVNGInputStream *input, const
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!input->isStructured())
     return false;
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   if (!stream)
     return false;
-  librevenge::RVNGInputStream *relStream = input->getSubStreamByName(getRelationshipsForTarget(name).c_str());
+  const RVNGInputStreamPtr_t relStream(input->getSubStreamByName(getRelationshipsForTarget(name).c_str()));
   input->seek(0, librevenge::RVNG_SEEK_SET);
-  VSDXRelationships rels(relStream);
-  if (relStream)
-    delete relStream;
+  VSDXRelationships rels(relStream.get());
   rels.rebaseTargets(getTargetBaseDirectory(name).c_str());
 
-  processXmlDocument(stream, rels);
+  processXmlDocument(stream.get(), rels);
 
-  delete stream;
   return true;
 }
 
@@ -219,19 +208,16 @@ bool libvisio::VSDXParser::parsePages(librevenge::RVNGInputStream *input, const
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!input->isStructured())
     return false;
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   if (!stream)
     return false;
-  librevenge::RVNGInputStream *relStream = input->getSubStreamByName(getRelationshipsForTarget(name).c_str());
+  const RVNGInputStreamPtr_t relStream(input->getSubStreamByName(getRelationshipsForTarget(name).c_str()));
   input->seek(0, librevenge::RVNG_SEEK_SET);
-  VSDXRelationships rels(relStream);
-  if (relStream)
-    delete relStream;
+  VSDXRelationships rels(relStream.get());
   rels.rebaseTargets(getTargetBaseDirectory(name).c_str());
 
-  processXmlDocument(stream, rels);
+  processXmlDocument(stream.get(), rels);
 
-  delete stream;
   return true;
 }
 
@@ -242,19 +228,16 @@ bool libvisio::VSDXParser::parsePage(librevenge::RVNGInputStream *input, const c
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!input->isStructured())
     return false;
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   if (!stream)
     return false;
-  librevenge::RVNGInputStream *relStream = input->getSubStreamByName(getRelationshipsForTarget(name).c_str());
+  const RVNGInputStreamPtr_t relStream(input->getSubStreamByName(getRelationshipsForTarget(name).c_str()));
   input->seek(0, librevenge::RVNG_SEEK_SET);
-  VSDXRelationships rels(relStream);
-  if (relStream)
-    delete relStream;
+  VSDXRelationships rels(relStream.get());
   rels.rebaseTargets(getTargetBaseDirectory(name).c_str());
 
-  processXmlDocument(stream, rels);
+  processXmlDocument(stream.get(), rels);
 
-  delete stream;
   return true;
 }
 
@@ -265,16 +248,50 @@ bool libvisio::VSDXParser::parseTheme(librevenge::RVNGInputStream *input, const
   input->seek(0, librevenge::RVNG_SEEK_SET);
   if (!input->isStructured())
     return false;
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   if (!stream)
     return false;
 
-  m_currentTheme.parse(stream);
+  m_currentTheme.parse(stream.get());
 
-  delete stream;
   return true;
 }
 
+void libvisio::VSDXParser::parseMetaData(librevenge::RVNGInputStream *input, libvisio::VSDXRelationships &rels) try
+{
+  if (!input)
+    return;
+  input->seek(0, librevenge::RVNG_SEEK_SET);
+  if (!input->isStructured())
+    return;
+
+  VSDXMetaData metaData;
+  const libvisio::VSDXRelationship *coreProp = rels.getRelationshipByType("http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties");
+  if (coreProp)
+  {
+    const RVNGInputStreamPtr_t stream(input->getSubStreamByName(coreProp->getTarget().c_str()));
+    if (stream)
+    {
+      metaData.parse(stream.get());
+    }
+  }
+
+  const libvisio::VSDXRelationship *extendedProp = rels.getRelationshipByType("http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties");
+  if (extendedProp)
+  {
+    const RVNGInputStreamPtr_t stream(input->getSubStreamByName(extendedProp->getTarget().c_str()));
+    if (stream)
+    {
+      metaData.parse(stream.get());
+    }
+  }
+  m_collector->collectMetaData(metaData.getMetaData());
+}
+catch (...)
+{
+  // Ignore any exceptions in metadata. They are not important enough to stop parsing.
+}
+
 void libvisio::VSDXParser::processXmlDocument(librevenge::RVNGInputStream *input, VSDXRelationships &rels)
 {
   if (!input)
@@ -282,57 +299,73 @@ void libvisio::VSDXParser::processXmlDocument(librevenge::RVNGInputStream *input
 
   m_rels = &rels;
 
-  xmlTextReaderPtr reader = xmlReaderForStream(input, 0, 0, XML_PARSE_NOBLANKS|XML_PARSE_NOENT|XML_PARSE_NONET);
+  XMLErrorWatcher watcher;
+
+  const boost::shared_ptr<xmlTextReader> reader(
+    xmlReaderForStream(input, 0, 0, XML_PARSE_NOBLANKS|XML_PARSE_NOENT|XML_PARSE_NONET, &watcher),
+    xmlFreeTextReader);
   if (!reader)
     return;
-  int ret = xmlTextReaderRead(reader);
-  while (1 == ret)
-  {
-    int tokenId = VSDXMLTokenMap::getTokenId(xmlTextReaderConstName(reader));
-    int tokenType = xmlTextReaderNodeType(reader);
 
-    switch (tokenId)
+  XMLErrorWatcher *oldWatcher = m_watcher;
+  try
+  {
+    m_watcher = &watcher;
+
+    int ret = xmlTextReaderRead(reader.get());
+    while (1 == ret && !watcher.isError())
     {
-    case XML_REL:
-      if (XML_READER_TYPE_ELEMENT == tokenType)
+      int tokenId = VSDXMLTokenMap::getTokenId(xmlTextReaderConstName(reader.get()));
+      int tokenType = xmlTextReaderNodeType(reader.get());
+
+      switch (tokenId)
       {
-        xmlChar *id = xmlTextReaderGetAttribute(reader, BAD_CAST("r:id"));
-        if (id)
+      case XML_REL:
+        if (XML_READER_TYPE_ELEMENT == tokenType)
         {
-          const VSDXRelationship *rel = rels.getRelationshipById((char *)id);
-          if (rel)
+          boost::shared_ptr<xmlChar> id(xmlTextReaderGetAttribute(reader.get(), BAD_CAST("r:id")), xmlFree);
+          if (id)
           {
-            std::string type = rel->getType();
-            if (type == "http://schemas.microsoft.com/visio/2010/relationships/master")
+            const VSDXRelationship *rel = rels.getRelationshipById((char *)id.get());
+            if (rel)
             {
-              m_currentDepth += xmlTextReaderDepth(reader);
-              parseMaster(m_input, rel->getTarget().c_str());
-              m_currentDepth -= xmlTextReaderDepth(reader);
+              std::string type = rel->getType();
+              if (type == "http://schemas.microsoft.com/visio/2010/relationships/master")
+              {
+                m_currentDepth += xmlTextReaderDepth(reader.get());
+                parseMaster(m_input, rel->getTarget().c_str());
+                m_currentDepth -= xmlTextReaderDepth(reader.get());
+              }
+              else if (type == "http://schemas.microsoft.com/visio/2010/relationships/page")
+              {
+                m_currentDepth += xmlTextReaderDepth(reader.get());
+                parsePage(m_input, rel->getTarget().c_str());
+                m_currentDepth -= xmlTextReaderDepth(reader.get());
+              }
+              else if (type == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image")
+              {
+                extractBinaryData(m_input, rel->getTarget().c_str());
+              }
+              else
+                processXmlNode(reader.get());
             }
-            else if (type == "http://schemas.microsoft.com/visio/2010/relationships/page")
-            {
-              m_currentDepth += xmlTextReaderDepth(reader);
-              parsePage(m_input, rel->getTarget().c_str());
-              m_currentDepth -= xmlTextReaderDepth(reader);
-            }
-            else if (type == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image")
-            {
-              extractBinaryData(m_input, rel->getTarget().c_str());
-            }
-            else
-              processXmlNode(reader);
           }
-          xmlFree(id);
         }
+        break;
+      default:
+        processXmlNode(reader.get());
+        break;
       }
-      break;
-    default:
-      processXmlNode(reader);
-      break;
+      ret = xmlTextReaderRead(reader.get());
     }
-    ret = xmlTextReaderRead(reader);
+
+    m_watcher = oldWatcher;
+  }
+  catch (...)
+  {
+    m_watcher = oldWatcher;
+    throw;
   }
-  xmlFreeTextReader(reader);
 }
 
 void libvisio::VSDXParser::processXmlNode(xmlTextReaderPtr reader)
@@ -407,7 +440,7 @@ void libvisio::VSDXParser::processXmlNode(xmlTextReaderPtr reader)
         readShapeProperties(reader);
       else
       {
-        if (m_isStencilStarted)
+        if (m_isStencilStarted && m_currentStencil)
           m_currentStencil->addStencilShape(m_shape.m_shapeId, m_shape);
         else
           _flushShape();
@@ -418,7 +451,7 @@ void libvisio::VSDXParser::processXmlNode(xmlTextReaderPtr reader)
     }
     else if (XML_READER_TYPE_END_ELEMENT == tokenType)
     {
-      if (m_isStencilStarted)
+      if (m_isStencilStarted && m_currentStencil)
         m_currentStencil->addStencilShape(m_shape.m_shapeId, m_shape);
       else
       {
@@ -479,7 +512,7 @@ void libvisio::VSDXParser::processXmlNode(xmlTextReaderPtr reader)
     {
       const xmlChar *name1 = xmlTextReaderConstName(reader);
       const xmlChar *value1 = xmlTextReaderConstValue(reader);
-      printf(" %s=\"%s\"", name1, value1);
+      fprintf(stderr, " %s=\"%s\"", name1, value1);
     }
   }
 
@@ -500,7 +533,7 @@ void libvisio::VSDXParser::extractBinaryData(librevenge::RVNGInputStream *input,
   if (!input || !input->isStructured())
     return;
   input->seek(0, librevenge::RVNG_SEEK_SET);
-  librevenge::RVNGInputStream *stream = input->getSubStreamByName(name);
+  const RVNGInputStreamPtr_t stream(input->getSubStreamByName(name));
   if (!stream)
     return;
   while (true)
@@ -512,7 +545,6 @@ void libvisio::VSDXParser::extractBinaryData(librevenge::RVNGInputStream *input,
     if (stream->isEnd())
       break;
   }
-  delete stream;
   VSD_DEBUG_MSG(("%s\n", m_currentBinaryData.getBase64Data().cstr()));
 }
 
@@ -621,9 +653,9 @@ void libvisio::VSDXParser::readPageSheetProperties(xmlTextReaderPtr reader)
       break;
     }
   }
-  while ((XML_PAGESHEET != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret);
+  while ((XML_PAGESHEET != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret && (!m_watcher || !m_watcher->isError()));
 
-  if (m_isStencilStarted)
+  if (m_isStencilStarted && m_currentStencil)
   {
     m_currentStencil->m_shadowOffsetX = shadowOffsetX;
     m_currentStencil->m_shadowOffsetY = shadowOffsetY;
@@ -664,7 +696,7 @@ void libvisio::VSDXParser::readFonts(xmlTextReaderPtr reader)
       ++idx;
     }
   }
-  while ((XML_FACENAMES != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret);
+  while ((XML_FACENAMES != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret && (!m_watcher || !m_watcher->isError()));
 }
 
 void libvisio::VSDXParser::readStyleProperties(xmlTextReaderPtr reader)
@@ -856,7 +888,7 @@ void libvisio::VSDXParser::readStyleProperties(xmlTextReaderPtr reader)
       break;
     }
   }
-  while ((XML_STYLESHEET != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret);
+  while ((XML_STYLESHEET != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret && (!m_watcher || !m_watcher->isError()));
 
 #if 0
   if (bgClrId < 0)
@@ -897,7 +929,7 @@ int libvisio::VSDXParser::getElementDepth(xmlTextReaderPtr reader)
 void libvisio::VSDXParser::readShapeProperties(xmlTextReaderPtr reader)
 {
   // Text block properties
-  long bgClrId = 0;
+  long bgClrId = -1;
 
   int ret = 1;
   int tokenId = XML_TOKEN_INVALID;
@@ -1006,6 +1038,38 @@ void libvisio::VSDXParser::readShapeProperties(xmlTextReaderPtr reader)
         ret = readDoubleData(m_shape.m_txtxform->angle, reader);
       }
       break;
+    case XML_BEGINX:
+      if (XML_READER_TYPE_ELEMENT == tokenType)
+      {
+        if (!m_shape.m_xform1d)
+          m_shape.m_xform1d = new XForm1D();
+        ret = readDoubleData(m_shape.m_xform1d->beginX, reader);
+      }
+      break;
+    case XML_BEGINY:
+      if (XML_READER_TYPE_ELEMENT == tokenType)
+      {
+        if (!m_shape.m_xform1d)
+          m_shape.m_xform1d = new XForm1D();
+        ret = readDoubleData(m_shape.m_xform1d->beginY, reader);
+      }
+      break;
+    case XML_ENDX:
+      if (XML_READER_TYPE_ELEMENT == tokenType)
+      {
+        if (!m_shape.m_xform1d)
+          m_shape.m_xform1d = new XForm1D();
+        ret = readDoubleData(m_shape.m_xform1d->endX, reader);
+      }
+      break;
+    case XML_ENDY:
+      if (XML_READER_TYPE_ELEMENT == tokenType)
+      {
+        if (!m_shape.m_xform1d)
+          m_shape.m_xform1d = new XForm1D();
+        ret = readDoubleData(m_shape.m_xform1d->endY, reader);
+      }
+      break;
     case XML_IMGOFFSETX:
       if (XML_READER_TYPE_ELEMENT == tokenType)
       {
@@ -1139,6 +1203,7 @@ void libvisio::VSDXParser::readShapeProperties(xmlTextReaderPtr reader)
             textBkgndColour = Colour(0xff, 0xff, 0xff, 0);
         }
         m_shape.m_textBlockStyle.textBkgndColour = textBkgndColour;
+        m_shape.m_textBlockStyle.isTextBkgndFilled = true;
       }
       break;
     case XML_DEFAULTTABSTOP:
@@ -1205,7 +1270,7 @@ void libvisio::VSDXParser::readShapeProperties(xmlTextReaderPtr reader)
       break;
     }
   }
-  while ((XML_SHAPES != tokenId) && (XML_SHAPE != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret);
+  while ((XML_SHAPES != tokenId) && (XML_SHAPE != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret && (!m_watcher || !m_watcher->isError()));
 
   if (1 == ret)
     processXmlNode(reader);
@@ -1229,7 +1294,7 @@ void libvisio::VSDXParser::readParagraph(xmlTextReaderPtr reader)
     if (XML_ROW == tokenId && XML_READER_TYPE_ELEMENT == tokenType)
       readParaIX(reader);
   }
-  while ((XML_SECTION != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret);
+  while ((XML_SECTION != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret && (!m_watcher || !m_watcher->isError()));
 }
 
 void libvisio::VSDXParser::readCharacter(xmlTextReaderPtr reader)
@@ -1250,7 +1315,7 @@ void libvisio::VSDXParser::readCharacter(xmlTextReaderPtr reader)
     if (XML_ROW == tokenId && XML_READER_TYPE_ELEMENT == tokenType)
       readCharIX(reader);
   }
-  while ((XML_SECTION != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret);
+  while ((XML_SECTION != tokenId || XML_READER_TYPE_END_ELEMENT != tokenType) && 1 == ret && (!m_watcher || !m_watcher->isError()));
 }
 
 void libvisio::VSDXParser::getBinaryData(xmlTextReaderPtr reader)
